package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.IntTag
import net.minecraft.nbt.ListTag
import net.minecraft.network.chat.Component
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.energy.IEnergyStorage
import net.minecraftforge.network.PacketDistributor
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.block.EnergyCounterBlock
import ru.dbotthepony.mc.otm.capability.*
import ru.dbotthepony.mc.otm.compat.mekanism.Mattery2MekanismEnergyWrapper
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.menu.EnergyCounterMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import java.lang.ref.WeakReference
import java.util.*

class EnergyCounterBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) : MatteryBlockEntity(MBlockEntities.ENERGY_COUNTER, p_155229_, p_155230_) {
	var passed by synchronizer.fraction()

	private val history = Array(10 * 20) { ImpreciseFraction.ZERO }
	internal var historyTick = 0

	var lastTick by synchronizer.fraction()
		internal set

	var ioLimit: ImpreciseFraction? = null

	fun resetStats() {
		historyTick = 0
		lastTick = ImpreciseFraction.ZERO
		passed = ImpreciseFraction.ZERO
		Arrays.fill(history, ImpreciseFraction.ZERO)
	}

	fun getHistory(ticks: Int): Array<ImpreciseFraction> {
		require(!(ticks < 1 || ticks >= history.size)) { "Invalid history length provided" }

		val history = Array(ticks) { ImpreciseFraction.ZERO }

		for (i in 0 until ticks) {
			var index = (historyTick - i) % this.history.size
			if (index < 0) index = this.history.size - 1
			history[i] = this.history[index]
		}

		return history
	}

	fun calcAverage(ticks: Int): ImpreciseFraction {
		return sumHistory(ticks) / ticks
	}

	fun sumHistory(ticks: Int): ImpreciseFraction {
		require(!(ticks < 1 || ticks >= history.size)) { "Invalid history length provided" }

		var value = ImpreciseFraction.ZERO

		for (i in 0 until ticks) {
			var index = (historyTick - i) % history.size
			if (index < 0) index = history.size - 1
			value += history[index]
		}

		return value
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt[PASSED_ENERGY_KEY] = passed.serializeNBT()

		val list = ListTag()
		nbt[POWER_HISTORY_KEY] = list
		nbt[POWER_HISTORY_POINTER_KEY] = historyTick

		for (num in history)
			list.add(num.serializeNBT())

		ioLimit?.let { nbt[IO_LIMIT_KEY] = it.serializeNBT() }
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		passed = nbt.getImpreciseFraction(PASSED_ENERGY_KEY)
		ioLimit = nbt.map(IO_LIMIT_KEY, ImpreciseFraction.Companion::deserializeNBT)

		nbt.ifHas(POWER_HISTORY_POINTER_KEY, IntTag::class.java) {
			historyTick = it.asInt
		}

		Arrays.fill(history, ImpreciseFraction.ZERO)

		for ((i, bytes) in nbt.getByteArrayList(POWER_HISTORY_KEY).withIndex()) {
			history[i] = ImpreciseFraction.deserializeNBT(bytes)
		}
	}

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return EnergyCounterMenu(containerID, inventory, this)
	}

	private val energyInput = EnergyCounterCap(true)
	private val energyOutput = EnergyCounterCap(false)

	private var inputCapability: LazyOptional<out IEnergyStorage> = LazyOptional.empty()
	private var outputCapability: LazyOptional<out IEnergyStorage> = LazyOptional.empty()

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)
		tickOnceServer(this::checkSurroundings)
	}

	private inner class EnergyCounterCap(val is_input: Boolean) : IMatteryEnergyStorage {
		override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			return extractEnergyInner(howMuch, simulate)
		}

		override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			if (is_input)
				return ImpreciseFraction.ZERO

			if (inputCapability.isPresent) {
				val it = inputCapability.resolve().get()

				val diff: ImpreciseFraction

				val ioLimit = ioLimit

				if (ioLimit != null) {
					diff = it.extractEnergy(howMuch.coerceAtMost(ioLimit), simulate)
				} else {
					diff = it.extractEnergy(howMuch, simulate)
				}

				if (!simulate) {
					passed += diff
					history[historyTick] += diff
				}

				return diff
			}

			return ImpreciseFraction.ZERO
		}

		override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			return receiveEnergyInner(howMuch, simulate)
		}

		override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			if (!is_input)
				return ImpreciseFraction.ZERO

			if (outputCapability.isPresent) {
				val it = outputCapability.resolve().get()

				val diff: ImpreciseFraction

				val ioLimit = ioLimit

				if (ioLimit != null) {
					diff = it.receiveEnergy(howMuch.coerceAtMost(ioLimit), simulate)
				} else {
					diff = it.receiveEnergy(howMuch, simulate)
				}

				if (!simulate) {
					passed += diff
					history[historyTick] += diff
				}

				return diff
			}

			return ImpreciseFraction.ZERO
		}

		override val batteryLevel: ImpreciseFraction
			get() {
				if (is_input) {
					if (outputCapability.isPresent) {
						val it = outputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.batteryLevel
						}

						return ImpreciseFraction(it.energyStored)
					}
				} else {
					if (inputCapability.isPresent) {
						val it = inputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.batteryLevel
						}

						return ImpreciseFraction(it.energyStored)
					}
				}

				return ImpreciseFraction.ZERO
			}

		override val maxBatteryLevel: ImpreciseFraction
			get() {
				if (is_input) {
					if (outputCapability.isPresent) {
						val it = outputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.maxBatteryLevel
						}

						return ImpreciseFraction(it.maxEnergyStored)
					}
				} else {
					if (inputCapability.isPresent) {
						val it = inputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.maxBatteryLevel
						}

						return ImpreciseFraction(it.maxEnergyStored)
					}
				}

				return ImpreciseFraction.ZERO
			}

		override val missingPower: ImpreciseFraction
			get() {
				if (is_input) {
					if (outputCapability.isPresent) {
						val it = outputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.missingPower
						}

						return ImpreciseFraction((it.maxEnergyStored - it.energyStored).coerceAtLeast(0))
					}
				} else {
					if (inputCapability.isPresent) {
						val it = inputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.missingPower
						}

						return ImpreciseFraction((it.maxEnergyStored - it.energyStored).coerceAtLeast(0))
					}
				}

				return ImpreciseFraction.ZERO
			}

		override fun canExtract() = !is_input
		override fun canReceive() = is_input
	}

	private var resolverInput = LazyOptional.of<IMatteryEnergyStorage> { energyInput }
	private var resolverOutput = LazyOptional.of<IMatteryEnergyStorage> { energyOutput }

	private var resolverInputMekanism = if (isMekanismLoaded) LazyOptional.of { Mattery2MekanismEnergyWrapper(energyInput) } else null
	private var resolverOutputMekanism = if (isMekanismLoaded) LazyOptional.of { Mattery2MekanismEnergyWrapper(energyInput) } else null

	private var valid = true

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		resolverInput.invalidate()
		resolverInputMekanism?.invalidate()
		resolverOutput.invalidate()
		resolverOutputMekanism?.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		resolverInput = LazyOptional.of { energyInput }
		resolverOutput = LazyOptional.of { energyOutput }

		if (isMekanismLoaded) {
			resolverInputMekanism = LazyOptional.of { Mattery2MekanismEnergyWrapper(energyInput) }
			resolverOutputMekanism = LazyOptional.of { Mattery2MekanismEnergyWrapper(energyInput) }
		}
	}

	@Suppress("deprecation")
	override fun setBlockState(new: BlockState) {
		val old = blockState
		super.setBlockState(new)

		if (new !== old && new.getValue(EnergyCounterBlock.INPUT_DIRECTION) != old.getValue(EnergyCounterBlock.INPUT_DIRECTION)) {
			resolverInput.invalidate()
			resolverInputMekanism?.invalidate()
			resolverOutput.invalidate()
			resolverOutputMekanism?.invalidate()

			resolverInput = LazyOptional.of { energyInput }
			resolverOutput = LazyOptional.of { energyOutput }

			if (isMekanismLoaded) {
				resolverInputMekanism = LazyOptional.of { Mattery2MekanismEnergyWrapper(energyInput) }
				resolverOutputMekanism = LazyOptional.of { Mattery2MekanismEnergyWrapper(energyInput) }
			}

			checkSurroundings()
		}
	}

	private fun getAndBind(
		level: Level,
		old: LazyOptional<out IEnergyStorage>,
		side: Direction
	): LazyOptional<out IEnergyStorage> {
		val ent = level.getBlockEntity(blockPos.offset(side.normal)) ?: return LazyOptional.empty()
		val resolve = ent.getEnergySided(-side)

		if (resolve !== old) {
			val weak = WeakReference(this)

			resolve.addListener {
				if (SERVER_IS_LIVE)
					weak.get()?.checkSurroundings()
			}

			return resolve
		}

		return old
	}

	fun checkSurroundings() {
		val level = level
		if (isRemoved || level !is ServerLevel) return

		inputCapability = getAndBind(
			level,
			inputCapability,
			blockState.getValue(EnergyCounterBlock.INPUT_DIRECTION)
		)

		outputCapability = getAndBind(
			level,
			outputCapability,
			-blockState.getValue(EnergyCounterBlock.INPUT_DIRECTION)
		)
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (side == null || isRemoved)
			return super.getCapability(cap, side)

		if (valid) {
			if (side == blockState.getValue(EnergyCounterBlock.INPUT_DIRECTION)) {
				if (cap == MatteryCapability.ENERGY || cap == ForgeCapabilities.ENERGY) {
					return resolverInput.cast()
				} else if (cap == MatteryCapability.MEKANISM_ENERGY) {
					return resolverInputMekanism!!.cast()
				}
			} else if (side == blockState.getValue(EnergyCounterBlock.INPUT_DIRECTION).opposite) {
				if (cap == MatteryCapability.ENERGY || cap == ForgeCapabilities.ENERGY) {
					return resolverOutput.cast()
				} else if (cap == MatteryCapability.MEKANISM_ENERGY) {
					return resolverOutputMekanism!!.cast()
				}
			}
		}

		return super.getCapability(cap, side)
	}

	fun tick() {
		lastTick = history[historyTick]
		historyTick = (historyTick + 1) % history.size
		history[historyTick] = ImpreciseFraction.ZERO
		synchronizeToPlayers()
	}

	fun clientTick() {
		historyTick = (historyTick + 1) % history.size
		history[historyTick] = lastTick
		passed += lastTick
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.energy_counter")
		const val PASSED_ENERGY_KEY = "passedEnergy"
		const val IO_LIMIT_KEY = "IOLimit"
		const val POWER_HISTORY_KEY = "passHistory"
		const val POWER_HISTORY_POINTER_KEY = "passHistoryPointer"
	}
}
