package ru.dbotthepony.mc.otm.compat.mekanism

import mekanism.api.Action
import mekanism.api.energy.IStrictEnergyHandler
import mekanism.api.math.FloatingLong
import mekanism.common.config.MekanismConfig
import net.minecraft.core.Direction
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.isMekanismLoaded
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import java.lang.ref.WeakReference
import java.util.WeakHashMap

private val mekanism2MtJ by lazy {
	try {
		val conf = MekanismConfig.general
		return@lazy conf.FORGE_CONVERSION_RATE.to.get().toImpreciseFraction()
	} catch(_: Throwable) {

	}

	return@lazy ImpreciseFraction.ONE
}

private val mtj2Mekanism by lazy {
	try {
		val conf = MekanismConfig.general
		return@lazy conf.FORGE_CONVERSION_RATE.from.get().toImpreciseFraction()
	} catch(_: Throwable) {

	}

	return@lazy ImpreciseFraction.ONE
}

class Mekanism2MatteryEnergyWrapper(private val power: IStrictEnergyHandler) : IMatteryEnergyStorage {
	override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		val action = when (simulate) {
			true -> Action.SIMULATE
			false -> Action.EXECUTE
		}

		return power.extractEnergy((howMuch * mtj2Mekanism).toFloatingLong(), action).toImpreciseFraction() * mekanism2MtJ
	}

	override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		return extractEnergyOuter(howMuch, simulate)
	}

	override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		val action = when (simulate) {
			true -> Action.SIMULATE
			false -> Action.EXECUTE
		}

		return howMuch - power.insertEnergy((howMuch * mtj2Mekanism).toFloatingLong(), action).toImpreciseFraction() * mekanism2MtJ
	}

	override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		return receiveEnergyOuter(howMuch, simulate)
	}

	override val batteryLevel: ImpreciseFraction
		get() {
			var sum = ImpreciseFraction.ZERO

			for (i in 0 until power.energyContainerCount) {
				sum += power.getEnergy(i).toImpreciseFraction()
			}

			return sum * mekanism2MtJ
		}

	override val maxBatteryLevel: ImpreciseFraction
		get() {
			var sum = ImpreciseFraction.ZERO

			for (i in 0 until power.energyContainerCount) {
				sum += power.getMaxEnergy(i).toImpreciseFraction()
			}

			return sum * mekanism2MtJ
		}
}

class Mattery2MekanismEnergyWrapper(private val power: IMatteryEnergyStorage) : IStrictEnergyHandler {
	override fun getEnergyContainerCount(): Int = 1

	override fun getEnergy(container: Int): FloatingLong {
		if (container != 0) {
			return FloatingLong.ZERO
		}

		return (power.batteryLevel * mtj2Mekanism).toFloatingLong()
	}

	override fun setEnergy(container: Int, value: FloatingLong) {
		// no op
	}

	override fun getMaxEnergy(container: Int): FloatingLong {
		if (container != 0) {
			return FloatingLong.ZERO
		}

		return (power.maxBatteryLevel * mtj2Mekanism).toFloatingLong()
	}

	override fun getNeededEnergy(container: Int): FloatingLong {
		if (container != 0) {
			return FloatingLong.ZERO
		}

		return (power.missingPower * mtj2Mekanism).toFloatingLong()
	}

	override fun insertEnergy(container: Int, howMuch: FloatingLong, action: Action): FloatingLong {
		val copy = howMuch.copy()
		return copy.minusEqual((power.receiveEnergyOuter(howMuch.toImpreciseFraction() * mekanism2MtJ, action.simulate()) * mtj2Mekanism).toFloatingLong())
	}

	override fun extractEnergy(container: Int, howMuch: FloatingLong, action: Action): FloatingLong {
		return (power.extractEnergyOuter(howMuch.toImpreciseFraction() * mekanism2MtJ, action.simulate()) * mtj2Mekanism).toFloatingLong()
	}
}

val ICapabilityProvider.mekanismEnergy: IMatteryEnergyStorage? get() {
	if (!isMekanismLoaded) {
		return null
	}

	val capability = getCapability(MatteryCapability.MEKANISM_ENERGY)

	if (!capability.isPresent) {
		return null
	}

	return Mekanism2MatteryEnergyWrapper(capability.orElseThrow(::IllegalStateException))
}

private val lazyCache by lazy { WeakHashMap<LazyOptional<IStrictEnergyHandler>, WeakReference<LazyOptional<Mekanism2MatteryEnergyWrapper>>>() }
private val lazyCacheDown by lazy { WeakHashMap<LazyOptional<IStrictEnergyHandler>, WeakReference<LazyOptional<Mekanism2MatteryEnergyWrapper>>>() }
private val lazyCacheUp by lazy { WeakHashMap<LazyOptional<IStrictEnergyHandler>, WeakReference<LazyOptional<Mekanism2MatteryEnergyWrapper>>>() }
private val lazyCacheSouth by lazy { WeakHashMap<LazyOptional<IStrictEnergyHandler>, WeakReference<LazyOptional<Mekanism2MatteryEnergyWrapper>>>() }
private val lazyCacheNorth by lazy { WeakHashMap<LazyOptional<IStrictEnergyHandler>, WeakReference<LazyOptional<Mekanism2MatteryEnergyWrapper>>>() }
private val lazyCacheEast by lazy { WeakHashMap<LazyOptional<IStrictEnergyHandler>, WeakReference<LazyOptional<Mekanism2MatteryEnergyWrapper>>>() }
private val lazyCacheWest by lazy { WeakHashMap<LazyOptional<IStrictEnergyHandler>, WeakReference<LazyOptional<Mekanism2MatteryEnergyWrapper>>>() }

fun ICapabilityProvider.getMekanismEnergySided(side: Direction? = null): LazyOptional<IMatteryEnergyStorage> {
	if (!isMekanismLoaded) {
		return LazyOptional.empty()
	}

	val lazyOptional = getCapability(MatteryCapability.MEKANISM_ENERGY, side)

	if (!lazyOptional.isPresent) {
		return LazyOptional.empty()
	}

	val cache = when (side) {
		Direction.DOWN -> lazyCacheDown
		Direction.UP -> lazyCacheUp
		Direction.NORTH -> lazyCacheNorth
		Direction.SOUTH -> lazyCacheSouth
		Direction.WEST -> lazyCacheWest
		Direction.EAST -> lazyCacheEast
		null -> lazyCache
	}

	val cached = cache[lazyOptional]?.get()

	if (cached != null) {
		return cached.cast()
	}

	val resolver = LazyOptional.of {
		Mekanism2MatteryEnergyWrapper(lazyOptional.orElseThrow(::IllegalStateException))
	}

	val ref = WeakReference(resolver)
	cache[lazyOptional] = ref

	lazyOptional.addListener {
		ref.get()?.invalidate()
	}

	return resolver.cast()
}

